(***********************************************************************)
(*                                                                     *)
(*          Run-time Code Generation Library for Objective Caml        *)
(*                                                                     *)
(*                   Dmitri Lomov (dsl@tepkom.ru)                      *)
(*              St-Petersburg State University, Russia                 *)
(*                                                                     *)
(***********************************************************************)    

(* $Id $ *)
open Opcodes
open Bytecode
open BytePrims

(* Label offsets *)
type label_offsets = 
  NotCreated of int list
| Created of int * (int list)

type label = Bytecode.label * label_offsets

type t_labels = (label, Bytecode.label) Env.t

let labels_hash_size = 32

let init_labels () =
  Env.init 
    (fun l -> l mod labels_hash_size)
    (fun (l,_) -> l)
    labels_hash_size

type t_label_state =
{
  labels : t_labels;
  next_label : Bytecode.label;
}

type t_externals_state =
{
  externals : Obj.t list;
  next_external : int;
}


type codegen =
{
 (* Generated bytecode *)
 mutable code : RawBytecode.t;

 (* Label definitions *)
 label_state : t_label_state;
                                                  
 (* Externals - last has number 0 and upwards *)
 externals_state : t_externals_state;

 (* Stack pointer *)
 mutable sp : int;  (* Grows upward *)
 
 (* Flags *)
 mutable prev_push : bool;

 (* Number of POPs to be done *)
 mutable pops : int;

 (* Generation flags *)
 unsafe : bool;
}


(* Initialization routines *)
let init () = RawBytecode.init ()
let uninit () = RawBytecode.uninit ()

let initial_bytecode_size = 256
let initial_bytecode_incr = 256

(* Starts generation *)
let start unsf =
  { 
    code = RawBytecode.create initial_bytecode_size;
    label_state =
    {
      labels = init_labels ();
      next_label = 0;
    };
    externals_state =
    {
      externals = [];
      next_external = 0;
    };
    sp = 0;
    prev_push = false;
    pops = 0;
    unsafe = unsf;
  }

exception UndefinedLabel of Bytecode.label

(* Shortcut for RawBytecode.put_data *)
let put = RawBytecode.put_data

(* Shortcut for RawBytecode.put_opcode *)
let put_opcode = RawBytecode.put_opcode

(* Shortcut for RawBytecode.put_data for codegen*)
let put_cg b cg =
  (cg.code <- put b cg.code; cg)

(* Place labels and correct code *)
let place_labels cg = 
  let place_label cg (l, s) =
    match s with
      Created(pos, refs) -> 
         let rec process_refs cg = function
	   [] -> ()
         | h::t ->
	     cg.code <- RawBytecode.put_data_at (pos-h) h cg.code;
	     process_refs cg t in
	 process_refs cg refs; cg 
    | (NotCreated _) -> 
         raise (UndefinedLabel l) in
  (Env.fold cg.label_state.labels cg place_label)
  
(* Perform final code transformations *)
let end_gen cg =
  let cg = 
    if cg.prev_push
    then 
     (cg.code <- (RawBytecode.put_opcode opPUSH cg.code);
      cg.prev_push <- false;
      cg)
    else if cg.pops > 0 
    then 
      (cg.code <- (put cg.pops (put_opcode opPOP cg.code));
       cg.pops <- 0;
       cg)
    else cg in
  place_labels cg

(* Finish code generation *)
let stop l cg =
  let cg =  end_gen cg in 
  let offset = 
    match Env.get cg.label_state.labels l with
      Some (_, Created(offset, _)) -> offset
    | _ -> raise (UndefinedLabel l) in
  RawBytecode.finish 
    cg.code 
    offset
    cg.externals_state.next_external 
    cg.externals_state.externals
  
(* Bytecode *)

exception DuplicatedLabel of int

(* Positions given label l *)
let create_label l cg =
  let ls = cg.label_state.labels and
      pos = RawBytecode.get_offset cg.code in
  match Env.get ls l with
    None ->
     { cg with
         label_state =
         { cg.label_state with labels = Env.add (l,(Created(pos, []))) ls } }
  | Some (_, (NotCreated r)) ->
     { cg with 
         label_state =
          { cg.label_state with
              labels = Env.add (l, (Created(pos, r))) ls } }
  | _ ->
      raise (DuplicatedLabel l) 
  
(* If PUSH instruction must be generated now,
 * generates it. *)
let gen_push cg next_is_push =
  if cg.prev_push
  then 
   (cg.code <- (RawBytecode.put_opcode opPUSH cg.code);
    cg.prev_push <- next_is_push;
    cg)
  else cg

(* Returns index of the comparison function *)
let get_compare_index = function
    LT -> get_lessthan_index ()
  | GT -> get_greaterthan_index ()
  | LE -> get_lessequal_index ()
  | GE -> get_greaterequal_index ()
  | EQ -> get_equal_index ()
  | NEQ -> get_notequal_index ()

exception NotImplemented of instruction

(* Generates given instruction without peep-hole optimiztaions *)
let instr_no_optimize i cg = 

  (* Generates given bytecode instruction with one operand *)
  let op1 opcode op sp_diff cg =
   (cg.code <- put op (put_opcode opcode cg.code);
    cg.sp <- cg.sp + sp_diff;
    cg)
      and
  (* Adds reference to given position for label l *)  
      add_ref labs l pos =
        match Env.get labs l with
          None -> 
            Env.add (l, NotCreated([pos])) labs
        | Some(_,NotCreated(r)) ->
	    Env.add (l, NotCreated(pos::r)) labs
        | Some(_,Created(at, r)) ->
	    Env.add (l, Created(at, pos::r)) labs
      in

  (* Generates given bytecode instruction with reference to label l *)
  let op1ref opcode l sp_diff cg =
        let c = RawBytecode.put_opcode opcode cg.code in
        let labs = add_ref 
                     cg.label_state.labels 
                     l 
                     (RawBytecode.get_offset c) in
        
        {(cg.code <- put 0 cg.code;
          cg.sp <- cg.sp + sp_diff;
          cg) with
            label_state = { cg.label_state with labels = labs } }
      in

  (* Body of instr_no_optimize *)
  let c = cg.code in
  match i with
    Ilabel l ->
      create_label l cg
  | IVequal ->
      cg.code <- put (get_equal_index ()) (put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg
  | IVcompare comparison ->
      cg.code <- put (get_compare_index comparison) 
	(put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg

  | Iassign n -> 
      op1 opASSIGN n 0 cg
  | Ipush_retaddr l -> 
      op1ref opPUSH_RETADDR l 3 cg
  | Iapply n ->
      let c =
        if(n >= 1) && (n <= 3)
        then put_opcode (opAPPLY+n) cg.code
        else put n (RawBytecode.put_opcode opAPPLY cg.code) in
      (cg.code <- c; 
      cg.sp <- cg.sp - n - 3;
      cg)
  | Iappterm(n, size) ->
      let c =
        if(n >= 1) && (n <= 3)
        then put_opcode (opAPPTERM+n) cg.code
        else put n (RawBytecode.put_opcode opAPPTERM cg.code) in
      (cg.code <- RawBytecode.put_data size c;
      cg.sp <- cg.sp - n + 1;
      cg) 
  | Ireturn n ->
      op1 opRETURN (n+cg.sp) 0 cg
  | Irestart -> 
      put_cg opRESTART cg
  | Igrab n -> 
      op1 opGRAB n 0 cg       
  | Iclosure n ->
      let cg = op1 opCLOSURE n (if n > 0 then 1-n else 0) cg in
      cg.code <- put 0 cg.code; cg
(* 
 *  now optimized
 * | Igetfield n ->
 *     op1 opGETFIELD n 0 cg
 * | Isetfield n ->
 *     op1 opSETFIELD n (-1) cg
 *)

  | Igetfloatfield n ->
      op1 opGETFLOATFIELD n 0 cg
  | Isetfloatfield n ->
      op1 opSETFLOATFIELD n (-1) cg
  | Imakeblock(size,tag) ->
      let cg = op1 opMAKEBLOCK size (-size+1) cg in
      cg.code <- put tag cg.code; cg
  | Imakefloatblock size ->
      op1 opMAKEFLOATBLOCK size (-size) cg
  | Iaddint | Isubint | Imulint | Idivint | Imodint |
    Iandint | Iorint  | Ixorint | Ilslint | Ilsrint |
    Iasrint | (Iintcomp _)  ->
      cg.code <- put_opcode (get_0ary_opcode i) cg.code;
      cg.sp <- cg.sp - 1;
      cg
  | Igetvectitem ->
      if cg.unsafe 
      then
	cg.code <- put_opcode opGETVECTITEM cg.code
      else
	cg.code <- put (get_array_get_addr_index ()) 
	    (put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg
  | Igetstringchar ->
      if cg.unsafe
      then 
	cg.code <- put_opcode opGETSTRINGCHAR cg.code
      else
	cg.code <- put (get_string_get_index ()) 
	    (put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg
  | Iboolnot | Inegint | Ivectlength | Istop | Igetmethod->
      cg.code <- put_opcode (get_0ary_opcode i) cg.code;
      cg
  | IVgetvectitem ->
      cg.code <- put (get_array_get_index()) (put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg
  | IVgetfloatitem ->
      cg.code <- 
	put (get_array_get_float_index()) (put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg
  | Isetvectitem ->
      if cg.unsafe
      then
	cg.code <- put_opcode opSETVECTITEM cg.code
      else
	cg.code <- 
	  put (get_array_set_addr_index ()) (put_opcode opC_CALL3 cg.code);
      cg.sp <- cg.sp - 2;
      cg
  | Isetstringchar ->
      if cg.unsafe 
      then 
	cg.code <- put_opcode opSETSTRINGCHAR cg.code
      else
	cg.code <- 
	  put (get_string_set_index ()) (put_opcode opC_CALL3 cg.code);
      cg.sp <- cg.sp - 2;
      cg
  | IVsetvectitem ->
      cg.code <- put (get_array_set_index()) (put_opcode opC_CALL3 cg.code);
      cg.sp <- cg.sp - 2;
      cg
  | IVsetfloatitem ->
      cg.code <- 
	put (get_array_set_float_index()) (put_opcode opC_CALL3 cg.code);
      cg.sp <- cg.sp - 2;
      cg
  | IVreg_closure ->
      cg.code <- put (RawBytecode.get_raw_register_closure_index ()) 
                   (put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg
 | IVrec_reg_closure ->
      cg.code <- put (RawBytecode.get_raw_rec_register_closure_index ()) 
           (put_opcode opC_CALL2 cg.code);
      cg.sp <- cg.sp - 1;
      cg
 | IVflop f ->
     cg.code <- put 
	 (match f with
	   Code.FO_ADD -> get_add_float_index ()
	 | Code.FO_SUB -> get_sub_float_index ()
	 | Code.FO_MUL -> get_mul_float_index ()
	 | Code.FO_DIV -> get_sub_float_index ()
	 | Code.FO_MOD -> get_fmod_float_index ()
	 | Code.FO_POWER -> get_power_float_index())
	 (put_opcode opC_CALL2 cg.code);
     cg.sp <- cg.sp - 1;
     cg
 | IVfloat_neg ->
     cg.code <- put (get_neg_float_index ()) cg.code;
     cg
 | Ioffsetint n ->
     op1 opOFFSETINT n 0 cg
 | Ioffsetref n ->
     op1 opOFFSETREF n 0 cg
 | Ibranch l ->
     op1ref opBRANCH l 0 cg
 | Ibranchif l ->
     op1ref opBRANCHIF l 0 cg
 | Ibranchifnot l ->
     op1ref opBRANCHIFNOT l 0 cg
 | Ibranchintcmp(cmp,const,l) ->
     let opcode =
       match cmp with
	   EQ -> opBEQ
	 | NEQ -> opBNEQ
	 | LE -> opBGEINT  (* VM compares const with accu *)
	 | GE -> opBLEINT  (* We compare accu with const *)
	 | LT -> opBGTINT 
	 | GT -> opBLTINT in
     let c = put const (put_opcode opcode cg.code) in
     let labs = add_ref cg.label_state.labels l
      (RawBytecode.get_offset c) in
     { (cg.code <- put 0 c; cg) with
	 label_state = { cg.label_state with labels = labs }
     }
     
 | _ -> 
     raise (NotImplemented i)

(* Generates PUSH+something short instructions *)
let gen_opt_push_short push_base no_push_base n cg =
  cg.code <- RawBytecode.put_opcode 
              ((if cg.prev_push then push_base else no_push_base)+n) 
              cg.code;
  cg.prev_push <- false;
  cg 

(* Generates PUSH+something long instructions *)
let gen_opt_push_long opcode_push opcode_no_push n cg =
  let c = RawBytecode.put_opcode
            (if cg.prev_push then opcode_push else opcode_no_push)
            cg.code in
  cg.code <- RawBytecode.put_data n c;
  cg.prev_push <- false;
  cg
      

let instr i cg =
  match i with
    Ipop n ->
      cg.pops <- cg.pops + n;
      cg.sp <- cg.sp - n;
      cg
  | _ ->
      let cg = 
	if cg.pops > 0
	then 
	  (cg.code <- (put cg.pops (put_opcode opPOP cg.code));
           cg.pops <- 0; cg)
        else cg in
      (match i with
	Ipush ->
	  let cg = gen_push cg true in 
	  cg.prev_push <- true; 
	  cg.sp <- cg.sp + 1;
	  cg  
	    
      | Iacc n -> 
	  if (n >= 0) && (n <= 7)
	  then 
            gen_opt_push_short opPUSHACC0 opACC0 n cg
	  else 
            gen_opt_push_long opPUSHACC opACC n cg  
      | Ienvacc n ->
	  if (n >= 1) && (n <= 4)
	  then 
            gen_opt_push_short (opPUSHENVACC1-1) (opENVACC1-1) n cg
	  else
            gen_opt_push_long opPUSHENVACC opENVACC n cg
      | Iconst n ->
	  if n >= 0 && n <= 3
	  then gen_opt_push_short opPUSHCONST0 opCONST0 n cg
	  else gen_opt_push_long opPUSHCONSTINT opCONSTINT n cg
      | Iatom n ->
	  if n = 0
	  then gen_opt_push_short opPUSHATOM0 opATOM0 n cg
	  else gen_opt_push_long opPUSHATOM opATOM n cg
      |	Igetfield n ->
	  if n >= 0 && n <= 3
	  then 
	    (cg.code <- put_opcode (opGETFIELD0 + n) cg.code; cg)
	  else 
	    (cg.code <- put n (put_opcode opGETFIELD cg.code); cg)
      |	Isetfield n ->
	  cg.code <-
	    if n >= 0 && n <= 3
	    then 
	      put_opcode (opSETFIELD0 + n) cg.code
	    else
	      put n (put_opcode opSETFIELD cg.code);
	  cg.sp <- cg.sp - 1;
	  cg
      |	Imakeblock(size,tag) ->
	  let code = 
	    if size >= 1 && size <= 3
	    then put_opcode (opMAKEBLOCK1+size-1) cg.code
	    else put size (put_opcode opMAKEBLOCK cg.code) in
	  cg.code <- put tag code;
	  cg.sp <- cg.sp - size + 1;
	  cg
      | _ -> instr_no_optimize i (gen_push cg false))
          
let getsp cg = cg.sp

let setsp sp cg = cg.sp <- sp; cg

let inline ptr len cg = 
  cg.code <- RawBytecode.copy_code ptr len cg.code;
  cg

(* Externals *)

let new_external o cg = 
  let rec find i = function 
    [] -> -1
  | h::t -> 
      if h == o
      then i
      else find (i-1) t in
  let e = cg.externals_state in
  match find 
          (e.next_external-1) 
          e.externals with
    -1 -> 
       let n = e.next_external in
       n, { cg with 
              externals_state =
              { e with 
                  next_external = n + 1; 
                  externals = o::e.externals } }
  | i -> i, cg
             

(* Label management *)
let newlabel cg =
  let l = cg.label_state.next_label in
  l, { cg with label_state = { cg.label_state with next_label = l + 1 } }


let is_unsafe cg = cg.unsafe

(*
;;
Printf.printf "xxx";
init ();
let cg = start () in
let st_l, cg = newlabel cg in
let cg = instr (Ilabel st_l) cg in 
let cg = instr (Iconst 1) cg in
let cg = instr (Ireturn 1) cg in
let handle = stop st_l cg in
let f : int -> int = 
  (let c = Obj.new_block Obj.closure_tag 3 in
  (Obj.magic (RawBytecode.register_closure handle c))) in
Printf.printf "And here!\n"; flush stdout;
Printf.printf "(f 1) = %d, (f 2) = %d\n" (f 1) (f 2);
uninit ();
Printf.printf "And here!!!\n"

*)


