(***********************************************************************)
(*                                                                     *)
(*          Run-time Code Generation Library for Objective Caml        *)
(*                                                                     *)
(*                   Dmitri Lomov (dsl@tepkom.ru)                      *)
(*              St-Petersburg State University, Russia                 *)
(*                                                                     *)
(***********************************************************************)    

(* $Id: inliner.ml,v 1.2 2002/03/01 11:40:22 mitya Exp $ *)
open Opcodes
open Bytecode
open GenHelper
open Bytegen
open Code


type generator_state = Bytegen.codegen * GenHelper.t_generation_state
type generator = 
    any_code -> generator_state -> data_tag -> bool -> generator_state
	
type code_pointer = RawBytecode.code_pointer
type code_area = code_pointer * int

external get_code_pointer : Obj.t -> code_pointer 
    = "raw_get_code_pointer" 
external test_instruction : code_pointer -> int -> bool 
    = "raw_test_instruction" 
external get_word : code_pointer -> int
    = "raw_get_word" 
external advance : code_pointer -> code_pointer
    = "raw_advance" 
external advance2 : code_pointer -> code_pointer
    = "raw_advance2"
external diff : code_pointer -> code_pointer -> int 
    = "raw_diff"


type record_access =
    { 
      ra_code_area : code_area;
      ra_is_float : bool;
    } 

type block_field =
    BFArg of int 
  | BFConst of int 
  | BFBlock of block
and block = 
    { block_size : int; 
      block_tag : int; 
      block_fields : block_field list; (* backwards *)
      block_float : bool;
    } 

type inlining =
    RecordAccess of record_access 
  | MakeBlock of block

let rec split_list n l =
  if n == 0
  then ([],l)
  else 
    match l with
      h::t ->
	let (l1,l2) = split_list (n-1) t in
	((h::l1), l2)
    | [] -> failwith "split_list" 

(***)
let rec print_field_list l =
  match l with
    h::t ->
      (match h with
	BFArg i -> Printf.eprintf "A%d" i
      | BFConst c -> Printf.eprintf "C%d" c
      | BFBlock b -> 
	  Printf.eprintf "["; 
	  print_field_list b.block_fields;
	  Printf.eprintf "]";);
      print_field_list t
  | [] -> ()

let print_preallocated s n_args =
  Printf.eprintf "{";
  for i = 0 to n_args do
    if BitSet.is s i 
    then Printf.eprintf "A%d" i
  done;
  Printf.eprintf "}"
(***)
  
let check_make_block start_ptr args n_args =
  let failure = None, args, n_args in
  (* checks whether instruction at ptr is in given range *)
  let rec instr_in_range ptr range_min range_max =
    if range_min > range_max
    then None
    else if test_instruction ptr range_min
    then Some range_min
    else instr_in_range ptr (range_min+1) range_max in

  (* test for first field creation *)
  let next_simple_field nfields ptr =
    (match instr_in_range ptr opCONST0 opCONSTINT with
      (* constant *)
      Some opcode ->
	if opcode == opCONSTINT
	then 
	  let ptr = advance ptr in
	  Some (advance ptr, (BFConst (get_word ptr)))
	else
	  Some (advance ptr, (BFConst (opcode - opCONST0))) 
    | None ->
	(* argument *)
	(match instr_in_range ptr opACC0 opACC with
	  Some opcode ->
	    let ptr, sp =
	      if opcode == opACC
	      then 
		let ptr = advance ptr in
		(advance ptr, get_word ptr)
	      else
		(advance ptr, opcode - opACC0) in
	    Some (ptr, BFArg(sp - nfields))
	| None ->
	    None
	) 
    ) in
    
  (* checks for the next field creation *)
  let next_pushy_field nfields ptr =
    if nfields == 0
    then next_simple_field 0 ptr
    else if test_instruction ptr opPUSH
    then 
      next_simple_field nfields (advance ptr)
    else      
      (match instr_in_range ptr opPUSHCONST0 opPUSHCONSTINT with
      (* constant *)
	Some opcode ->
	  if opcode == opPUSHCONSTINT
	  then 
	    let ptr = advance ptr in
	    Some (advance ptr, (BFConst (get_word ptr)))
	  else
	    Some (advance ptr, (BFConst (opcode - opPUSHCONST0))) 
      | None ->
	(* argument *)
	  (match instr_in_range ptr opPUSHACC0 opPUSHACC with
	    Some opcode ->
	      let ptr, sp =
		if opcode == opPUSHACC
		then 
		let ptr = advance ptr in
		(advance ptr, get_word ptr)
		else
		(advance ptr, opcode - opPUSHACC0) in
	    Some (ptr, BFArg(sp - nfields))
	  | None ->
	      None
	  )
      ) in

  let rec build_block nfields field_list ptr =
    (match next_pushy_field nfields ptr with
      Some (ptr, fld) ->
	build_block (nfields+1) (fld::field_list) ptr
    | None ->
	(match instr_in_range ptr opMAKEBLOCK opMAKEBLOCK3 with
	  Some opcode ->
	    let (ptr,size) =
	      if opcode == opMAKEBLOCK
	      then 
		let ptr = advance ptr in
		(advance ptr, get_word ptr)
	      else (advance ptr, opcode - opMAKEBLOCK) in
	    let tag = get_word ptr in
	    if test_instruction (advance ptr) opRETURN
	    then 
	      Some { block_size = size; block_tag = tag;
		     block_fields = List.rev field_list; block_float = false }
	    else
	      let 
		 (subblock_fields,block_fields) = split_list size field_list in
	      let subblock =
		{ block_size = size; block_tag = tag;
		  block_fields = List.rev subblock_fields; 
		  block_float = false } in
	      build_block 
		(nfields-size+1) 
		((BFBlock subblock)::block_fields) 
		(advance ptr) 
	| None ->
	    None
	)
    ) in
  if test_instruction start_ptr opGRAB
  then
    let ptr = advance start_ptr in
    if get_word ptr == n_args - 1
    then
      (match build_block 0 [] (advance ptr) with
	Some block ->
	  Some(MakeBlock block), [], 0
      |	None -> failure
      )
    else
      failure
  else
    (match build_block 0 [] start_ptr with
      Some block ->
	Some(MakeBlock block), [], 0
    | None -> failure
    )

let check closure args n_args =
  let failure = (None, args, n_args) in
  let rec check_get_field ptr =
    if (test_instruction ptr opGETFIELD0) or
      (test_instruction ptr opGETFIELD1) or
      (test_instruction ptr opGETFIELD2) or
      (test_instruction ptr opGETFIELD3) 
    then check_get_field (advance ptr)
    else if (test_instruction ptr opGETFIELD)
    then check_get_field (advance2 ptr)
    else ptr 
  in

  let start_ptr = get_code_pointer closure in
  if not (test_instruction start_ptr opACC0)
  then 
    check_make_block start_ptr args n_args
  else
    let copy_area_start = (advance start_ptr) in
    let ptr = check_get_field copy_area_start in
    (* checking for blocks of length 1 *)
    if test_instruction ptr opMAKEBLOCK1
    then
      let ptr = advance ptr in
      let tag = get_word ptr in
      if test_instruction (advance ptr) opRETURN 
      then 
	(match args with
	  [_] ->
	    (Some (
	     MakeBlock
	       { block_size = 1; block_tag = tag; 
		 block_fields = [(BFArg 0)]; block_float = false}),
	     [],0)
	| _ -> assert false)
      else failure
    else
      (* checking for field access *)
      let ptr, is_float =
	if test_instruction ptr opGETFLOATFIELD
	then (advance2 ptr), true
	else ptr, false in
      if (test_instruction ptr opRETURN)
      then 
	match args with
	  record::others ->
	    (Some (RecordAccess 
		     { ra_code_area = 
		       copy_area_start, (diff copy_area_start ptr);
		       ra_is_float = is_float }),
	     others, n_args-1)
      else
	failure
	  
      

let analyze cd args n_args =
  match Obj.magic cd with
    Quote closure ->
      assert (Obj.tag closure == Obj.closure_tag);
      check closure args n_args
  | _ ->
      (None, args, n_args) 

let remake_block block args n_args gen state =
  (* positions of arguments in the stack *)
  let positions = Array.make n_args 0xBADF00D in
  (* generates all the code necessary to make block *)
  let rec create_block 
      { block_size = block_size; block_tag = block_tag; 
	block_fields = block_fields } start_pos state =
    let push_field (pos,((cg,gs) as state)) field =
      (pos+1,
       let cg = 
	 if pos > start_pos
	 then Bytegen.instr Ipush cg
	 else cg in
       (match field with
	 BFConst const ->
	   let cg = Bytegen.instr (Iconst const) cg in
	   (cg, { gs with last_tag = unboxed_tag })
       | BFArg index ->
	   let old_pos = positions.(index) in
	    (if  old_pos == 0xBADF00D
	    then
	      let arg_code = List.nth args index in
	      positions.(index) <- pos;
	      gen arg_code (cg,gs) unknown_tag false 
	    else
	      ((instr (Iacc (pos - old_pos - 1)) cg), gs)
	   )
       | BFBlock block ->
	   create_block block pos state
       | _ -> assert false
       )
      ) in
    let (_,(cg,gs)) = 
      List.fold_left push_field (start_pos,state) block_fields  in
    ((instr (Imakeblock(block_size,block_tag)) cg),
     { gs with last_tag = block_tag }) in

  (* find specials in the list *)
  let existing = BitSet.make n_args 
  and killed = BitSet.make n_args
  and preallocated = BitSet.make n_args in
  let rec specials block_fields =
    let here = BitSet.make n_args in
    let process_field field =
      (match field with
	BFConst _ -> ()
      |	BFArg i ->
	  if BitSet.is killed i
	  then BitSet.on preallocated i
	  else if not (BitSet.is existing i)
	  then (BitSet.on here i; BitSet.on existing i)
      |	BFBlock { block_fields = block_fields } ->
	  specials block_fields
      ) in
    List.iter process_field block_fields;
    BitSet.union killed here;
    BitSet.sub existing here in

  let preallocate_args s state =
    let rec count i res =
      if i ==  n_args 
      then res
      else if BitSet.is s i 
      then count (i+1) (res+1)
      else count (i+1) res in
    let npre = count 0 0 in
    let rec iter i pos ((cg,gs)as state) =
      if i == n_args
      then state
      else if BitSet.is s i
      then 
	let cd = List.nth args i in
	let (cg,gs) = gen cd state unknown_tag false in
	let cg = instr Ipush cg in
	positions.(i) <- pos;
	iter (i+1) (pos+1) (cg,gs)
      else iter (i+1) pos state in
    iter 0 (-npre) state in	
  specials block.block_fields;
  create_block block 0 (preallocate_args preallocated state) 

let inline inlining args n_args gen state last_result =
  match inlining with
    RecordAccess ra ->
      (match args with
	arg::_ ->
	  let (cg,gs) = gen arg state unknown_tag last_result in
	  let cg = 
	    let (ptr,len) = ra.ra_code_area in
	    Bytegen.inline ptr len cg and
	      gs =
	    if ra.ra_is_float 
	    then (gs.last_tag <- Obj.double_tag; gs)
	    else gs
	  in
	  (cg,gs)
      |	_ -> assert false)
  | MakeBlock block ->
      remake_block block args n_args gen state
  | _ -> assert false
    
(*
 * $Log: inliner.ml,v $
 * Revision 1.2  2002/03/01 11:40:22  mitya
 * In check_make_block, we no longer assume that number of
 * arguments in "block-making function" msut be more than 1.
 *
 *)
