(**************************************************************
 *
 * Syntax extensions for runtime code generation
 * 
 **************************************************************)
 
(* $Id: dml.ml4,v 1.8 2002/04/29 11:00:16 mitya Exp $ *) 
 
open Pcaml

(* simple (var or _) or tuple pattern
 * No other patterns in DML yet *)
type pattern =
    SimplePatt of string option
  | TuplePatt of string option list

(* Generates a pair of types  (tuple of codes, code of tuple) 
 *  where tuple is of length n
 *)
let gen_tuple_type loc n = 
  let rec aux n code_tuple tuple_code =
    if n == 0
    then code_tuple, tuple_code
    else
      let id = <:ctyp<'$"a"^(string_of_int n)$ >> in
      aux (n-1) (<:ctyp< Rtcg.code $id$ >>::code_tuple) (id::tuple_code) in
  aux n [] []

exception IdentifierUsedMoreThanOnceInPatt of string

(* Verifes that each id appears once in tuple pattern *)
let rec verify_tuple_patt patt id_list =
  match patt with
    [] -> None
  | h::t ->
      (match h with
	None -> verify_tuple_patt t id_list
      |	Some id ->
	  if List.exists ((=) id) id_list
	  then Some id
	  else verify_tuple_patt t (id::id_list))


(* Generates a pattern-matching code for tuple pattern *) 
(* Returns: (string option, expr,ctyp list) - (identifier, expression) *)
let gen_tuple_patts loc body patt =
  let rec slice_let index tuple_id cd typ = function 
      [] -> tuple_id, cd, List.rev typ
    | h::t ->
	let typ_id = <:ctyp< '$"a"^(string_of_int index)$>> in
	let new_typ = typ_id::typ in
	(match h with
	  None ->  (* This is _ *)
	    slice_let (index+1) tuple_id cd new_typ t
	| Some id -> (* Identifier *)
	    let tuple_id, s =
	      (match tuple_id with
		None -> Some id, id
	      | Some s -> tuple_id, s) in	   
	    let cd = 
	      <:expr< Rtcg.let' 
		(Obj.magic 
		   (Rtcg.unsafe_slice (Obj.repr $lid:s$) 
		      $int:(string_of_int index)$ ))
		(fun ($lid:id$ : Rtcg.code $typ_id$) -> $cd$) >> in
	    slice_let (index+1) tuple_id cd new_typ t) in
  match verify_tuple_patt patt [] with
    None -> slice_let 0 None body [] patt
  | Some id ->  (* error *)
      Stdpp.raise_with_loc loc (IdentifierUsedMoreThanOnceInPatt id)
    
(* Generates a pattern-matching code for DML pattern *)
let build_patt_fun loc body = function 
    SimplePatt (Some id) ->
      <:expr< fun $lid:id$ -> $body$ >> 
  | SimplePatt None ->
      <:expr< fun _ -> $body$ >> 
  | TuplePatt l ->
      let tuple_id, new_body, typ = gen_tuple_patts loc body l in
      (match tuple_id with
	None ->  <:expr< fun _ -> $body$ >>
      | Some s -> <:expr< fun ( $lid:s$ : Rtcg.code ($list:typ$))
	  -> $new_body$ >>)
 
(* Builds an abstraction from list of input DML patterns *)
(* and functuion body *)
let build_abstr loc vars body = 
  let rec build_abstr_aux vars = 
    match vars with
      [] -> body
    | h::t -> 
	<:expr< Rtcg.abstr $build_patt_fun loc (build_abstr_aux t) h$ >>
  in
  build_abstr_aux vars

(* Generates a tuple constructor with type annotations *)
let gen_tuple loc t =
  let list_from, list_to = gen_tuple_type loc (List.length t) in
  <:expr< 
    (let f : ( $list:list_from$ ) -> Rtcg.code ( $list:list_to$ ) = 
      fun x -> Obj.magic (Rtcg.unsafe_tuple (Obj.repr x)) in f) ($list:t$)  >>


(* 
 * Abstract data types
 *)

(* Constructor expression *)
type constr_arg =
    CAExpr of MLast.expr
  | CATuple of constr_arg list
  | CAConstr of string * constr_arg * MLast.loc
  | CAConstrNull of string * MLast.loc

(* Printer for constr_arg *)
let rec print_constr_arg = function
    CAExpr _ -> 
      Printf.eprintf "E"
  | CATuple l ->
      Printf.eprintf "[";
      List.iter print_constr_arg l;
      Printf.eprintf "]"
  | CAConstr(s,c,_) ->
      Printf.eprintf "(%s" s; 
      print_constr_arg c;
      Printf.eprintf ")"
  | CAConstrNull(s,_) ->
      Printf.eprintf "(%s)" s

(* Generates a function application from constr_arg *)
let fun_of_constr_arg ca =
  let rec expr_of_constr_arg 
      (ca : constr_arg) (id : int) (args : MLast.expr list) =
    (match ca with
      CAExpr e ->
	let loc = MLast.loc_of_expr e in
	let s = "x"^(string_of_int id) in
	([<:expr< $lid: s$>>], (id+1), (e::args))
    | CAConstrNull(s,loc) ->
	([<:expr< $uid: s$ >>], id, args)
    | CAConstr(s,c,loc) ->
	let (elist,id,args) = expr_of_constr_arg c id args in
	let application =
	  List.fold_left 
	    (fun e arg -> <:expr< $e$ $arg$ >>) 
	    <:expr< $uid:s$ >> elist in
	([application],id,args)
    | CATuple calist ->
	List.fold_left
	  (fun (elist,id,args) ca ->
	    let (elist',id,args) = expr_of_constr_arg ca id args in
	    ((elist @ elist'), id, args))
	  ([],id,args) calist 
    ) in
  let (body,maxid,args) = expr_of_constr_arg ca 0 [] in
  let args = List.rev args in
  let body =
    (match body with
      [body] -> body
    | _ -> assert false
    ) in
  let rec iter id e =
    if id = maxid
    then e
    else 
      let loc = MLast.loc_of_expr e in
      let s = "x"^(string_of_int id) in
      iter (id+1) <:expr< fun $lid:s$ -> $e$ >> in
  let fexpr = iter 0 body in
  let loc = MLast.loc_of_expr fexpr in
  List.fold_left 
    (fun e arg ->
      <:expr< Rtcg.apply $e$ $arg$ >>)
    <:expr< Rtcg.quote $fexpr$ >>
    args

type qualified_id = (string list) * string

let expr_of_value_path loc (modules,lid) =
  List.fold_right 
    (fun uid expr -> <:expr< $uid: uid$ . $expr$ >>)
    modules <:expr< $lid: lid$ >>

let patt_of_value_path loc (modules,lid) =
  List.fold_right 
    (fun uid expr -> <:patt< $uid: uid$ . $expr$ >>)
    modules <:patt< $lid: lid$ >>


(*
 * Pattern-matching builder
 *)

let rec remove_dupes = function
    [] -> []
  | h::t ->
      let t' = remove_dupes t in
      if List.mem h t'
      then t'
      else h::t'
  

let vars_of_patt patt =
  let rec vars_of_patt vars patt =
    match patt with
	<:patt< $lid:s$ >> ->         (* "Variable" *)
	  s::vars
      | <:patt< ($p1$ as $p2$) >>     (* "as" patterns *)
      | <:patt< $p1$ $p2$ >>          (* constructor application *)
      | <:patt< $p1$ | $p2$ >> ->     (* "or" patterns *)
	  vars_of_patt (vars_of_patt vars p1) p2
      | <:patt< ( $list: pl$ ) >> ->  (* tuples *)
	  List.fold_left vars_of_patt vars pl 
      | <:patt< { $list:ppl$ } >> ->       (* records: (label, pattern) list *)
	  List.fold_left 
	    (fun vars (_,p) -> vars_of_patt vars p) vars ppl
      | <:patt< ($p$ : $t$) >> -> (* type constraint *)
	  vars_of_patt vars p
      | _ -> vars in
  let vars = vars_of_patt [] patt in
  remove_dupes vars
	  
let build_match loc cond cases =

  (* Generates function names for match cases, such
   * that they are not in a given names_list
   *)
  let gen_fun_name =
    let i = ref 0 in
    fun names_list ->
      let rec try_next () =
	let id = "f"^(string_of_int !i) in
	i := !i+1;
	if List.mem id names_list
	then try_next ()
	else id in
      try_next () in

  (* Builds a tuple 
   *   (fun_id, (arg_list, expr), (patt, guard, fun_appl)) 
   * from match case tuple (patt, guard, expr)
   *)
  let fun_appl_of_case (patt, w, e) =
    let vars = vars_of_patt patt in
    let fid = gen_fun_name vars in
    let fun_appl =
      let loc = MLast.loc_of_patt patt in
      match vars with
	  [] -> <:expr< $lid:fid$ () >>
	| _ ->     
	    List.fold_left 
	      (fun e v -> <:expr< $e$ $lid:v$ >>) 
               <:expr< $lid:fid$ >> vars in
    (fid, (vars, e), (patt,w,fun_appl)) in

  let (fids,decls,cases') =
    List.fold_right
      (fun pwe (fl,dl,cl) ->
	 let (f,d,c) = fun_appl_of_case pwe in
	 ((f::fl),(d::dl),(c::cl))) 
      cases ([],[],[]) in

  (* Function performing a matching *)
  let match_fun = 
    let match_body = 
      <:expr< match e with [ $list: cases'$ ] >> in
    let match_fun_f =
      List.fold_right
	(fun fid e -> <:expr< fun $lid:fid$ -> $e$ >>)
	fids match_body in
    <:expr< fun e -> $match_fun_f$ >> in
  
  (* Generates a DML abstraction from (arg_list, expr) tuple *)
  let abstr_of_decl (args,expr) =
    match args with
	[] -> <:expr< Rtcg.abstr (fun _ -> $expr$) >>
      | _ ->
	  List.fold_right
	  (fun arg e -> 
	     <:expr< Rtcg.abstr (fun $lid:arg$ -> $e$) >>)
	  args expr in

  (* Application of matching function *)
  let match_appl =
    List.fold_left
      (fun e decl ->
	 let decl_abstr = abstr_of_decl decl in
	 <:expr< Rtcg.apply $e$ $decl_abstr$ >>) 
      <:expr< Rtcg.apply (Rtcg.quote $match_fun$) $cond$ >>
      decls in
  match_appl

let gen_list_creation loc el =
  let len = List.length el in
  let rec iter i idl =
    if i = len
    then idl
    else 
      let id = "x"^(string_of_int i) in
      iter (i+1) (id :: idl) in
  let idl = iter 0 [] in
  let list_fun =
    let list_creation =
      List.fold_left 
	(fun e id -> <:expr< [$lid:id$::$e$] >>)
      <:expr< [] >> idl in
    List.fold_left 
      (fun e id -> <:expr< fun $lid:id$ -> $e$ >>)
      list_creation idl in
  let list_fun_apply =
    List.fold_left
      (fun appl arg -> <:expr< Rtcg.apply $appl$ $arg$>>) 
    <:expr< Rtcg.quote $list_fun$ >> el in
  list_fun_apply

(*
 * Record creation
 *)
let gen_record_creation loc base lab_list =
  let len = List.length lab_list in
  let lab_expr_list =
    let gen_lab =
      let i = ref 0 in
      fun (qid,e) ->
	let loc = MLast.loc_of_expr e in
	let s = "x"^(string_of_int !i) in
	i := !i + 1;
	((patt_of_value_path loc qid), <:expr< $lid:s$ >>) in
    List.map gen_lab lab_list in
  let body = 
    match base with
	None -> <:expr< { $list:lab_expr_list$ } >>
      | Some _ -> <:expr< { (r) with $list:lab_expr_list$ } >> in
  let fcode =
    let rec iter n res =
      if n < 0
      then res
      else 
	let loc = MLast.loc_of_expr res in
	let s = "x"^(string_of_int n) in
	iter (n-1) <:expr< fun $lid:s$ -> $res$ >> in
    let fcode' = iter (len-1) body in
    (match base with
	None -> fcode'
      | Some _ ->
	  let loc = MLast.loc_of_expr fcode' in
	  <:expr< fun r -> $fcode'$ >>) in
  let init =
    match base with
	None -> <:expr< Rtcg.quote $fcode$ >>
      | Some e ->
	  <:expr< Rtcg.apply (Rtcg.quote $fcode$) $e$ >> in
  List.fold_left 
    (fun e (_,e') -> <:expr< Rtcg.apply $e$ $e'$ >>)
    init
    lab_list

let test_label_eq =
  let stream_peek_nth n strm =
    let rec loop n = function
	[] -> None
      | [x] -> if n == 1 then Some x else None
      | _ :: l -> loop (n - 1) l in
  loop n (Stream.npeek n strm) in    
  let rec test lev strm =
    match stream_peek_nth lev strm with
	Some (("UIDENT", _) | ("LIDENT", _) | ("", ".")) ->
          test (lev + 1) strm
      | Some ("", "=") -> ()
      | _ -> raise Stream.Failure 
  in
  Grammar.Entry.of_parser gram "test_label_eq" (test 1) 

(* cdexpr is a global grammar entry and can be extended by users *)
let cdexpr = Grammar.Entry.create gram "cdexpr"

EXTEND
    GLOBAL: patt expr cdexpr;

expr: 
    LEVEL "simple" [[ "{<"; e = cdexpr; ">}" -> e ]];

expr:
    LEVEL "apply" [[ "ref"; e = expr -> <:expr< Pervasives.ref $e$ >> ]];

cdpatt: 
  [
    [ l = LIST1 [n = LIDENT -> Some n | "_" -> None]
      SEP "," ->
	match l with
	  [e] -> SimplePatt e
	| _ -> TuplePatt l
    ]
  ];  	

cdexpr:
  [
      "top" RIGHTA
      [
        e1 = cdexpr; ";"; e2 = cdexpr ->
	  <:expr< Rtcg.seq $e1$ $e2$ >> 
      ] 
    | "expr1" 
      [
        "fun"; vars = LIST1 cdpatt; "->";  body = cdexpr ->  
	  build_abstr loc vars body

      | "fix"; var = LIDENT; "->"; body = cdexpr ->
	  <:expr< Rtcg.fix (fun $lid:var$ -> $body$)>>

      |	"let"; vars = LIST1 cdpatt; "="; arg = cdexpr; "in"; 
	  body = cdexpr ->
	    let v1::v = vars in
	    let abstr = build_abstr loc v arg in
	    if
	      match v1 with
		SimplePatt (Some s) ->
		  (match body with 
		    <:expr< $lid:k$ >> when k = s -> false
		  | _ -> true)
	      |	 _ -> true
	    then
	      <:expr< Rtcg.let' $abstr$ $build_patt_fun loc body v1$ >>
	    else
	      abstr

      | "match"; cond = SELF; "with"; OPT "|";
	  l = LIST1 cd_match_case SEP "|" ->
	    build_match loc cond l


      |	"let"; "rec";  
	   f = LIDENT; v = LIST0 cdpatt;  "="; arg = cdexpr; "in"; 
	   body = cdexpr ->
	    let abstr = build_abstr loc v arg in
	    if 
	      match body with
		<:expr< $lid:k$ >> -> k = f
	      |	_ -> false
	    then
	      <:expr< Rtcg.fix (fun $lid:f$ -> $abstr$) >>
            else	     
	      <:expr< Rtcg.let' 
		(Rtcg.fix (fun $lid:f$ -> $abstr$))
		(fun $lid:f$ -> $body$) >>

      | "if"; cond = cdexpr; "then"; th = cdexpr LEVEL "expr1";
	  "else"; el = cdexpr LEVEL "expr1" ->
	  <:expr< Rtcg.if' $cond$ $th$ $el$ >>

      |	"for"; 
	    var = LIDENT; "="; 
	    low = cdexpr;
	    d = 
	      [ "to" -> <:expr<Rtcg.DTo>> | "downto" -> <:expr<Rtcg.DDownTo>>];
	    hi = cdexpr; "do"; body = cdexpr; "done" ->
	   <:expr< Rtcg.for' $low$ $d$ $hi$ (fun $lid:var$ -> $body$) >>
      |	"while"; cond = cdexpr; "do"; body = cdexpr; "done" ->
	  <:expr< Rtcg.while' $cond$ $body$ >>
      ]

    | "," LEFTA
      [ e = SELF; ","; el = LIST1 NEXT SEP "," ->
	let t = e::el in
	match t with
	  [e1;e2] -> <:expr<Rtcg.tuple2 $e1$ $e2$>>
	| [e1;e2;e3] -> <:expr<Rtcg.tuple3 $e1$ $e2$ $e3$>>
	| _ -> gen_tuple loc t
      ]
	
    | ":=" NONA
      [ e1 = cdexpr; ":="; e2 = cdexpr ->
	<:expr< Rtcg.ref_set $e1$ $e2$ >>
      ]	
    		
    | "||" RIGHTA
      [ e1 = cdexpr; "||"; e2 = cdexpr -> 
	<:expr< Rtcg.bor $e1$ $e2$ >>
      ]

    | "&&" RIGHTA
      [ e1 = cdexpr; "&&"; e2 = cdexpr -> 
	<:expr< Rtcg.band $e1$ $e2$ >>
      ] 

    | "not" NONA
      [ "not"; e = cdexpr ->
	<:expr< Rtcg.bnot $e$ >> ] 
 
    | "<" LEFTA
      [
        e1 = cdexpr; "="; e2 = cdexpr ->
	  <:expr<Rtcg.equal $e1$ $e2$ >>
      | e1 = cdexpr; "<>"; e2 = cdexpr ->
	  <:expr<Rtcg.bnot (Rtcg.equal $e1$ $e2$) >>
      |	e1 = cdexpr; "=="; e2 = cdexpr ->
	  <:expr<Rtcg.eq $e1$ $e2$ >>
      |	e1 = cdexpr; "!="; e2 = cdexpr ->
	  <:expr<Rtcg.neq $e1$ $e2$ >>
      |	e1 = cdexpr;
	f = [ "<" -> <:expr< Rtcg.LT >>
            | ">" -> <:expr< Rtcg.GT >>
	    | "<=" -> <:expr< Rtcg.LE >>
	    | ">=" -> <:expr< Rtcg.GE >> ];
	e2 = cdexpr ->
	  <:expr< Rtcg.compare $e1$ $f$ $e2$ >> 
      ]  
    | RIGHTA
      [ e1 = SELF; "::"; e2 = SELF ->
	  <:expr< Rtcg.apply 
	     (Rtcg.apply (Rtcg.quote (fun x1 x2 -> [x1::x2])) $e1$) $e2$ >>
      ]	
    | "+" LEFTA
      [ e1 = cdexpr;
        f = [ "+" -> <:expr< Rtcg.IOpADD >> 
            | "-" -> <:expr< Rtcg.IOpSUB >> ];
	e2 = cdexpr ->
	  <:expr< Rtcg.int_bin $e1$ $f$ $e2$ >>
      | e1 = cdexpr;
	f = [ "+." -> <:expr< Rtcg.FLOP_ADD >>
            | "-." -> <:expr< Rtcg.FLOP_SUB >> ];
	e2 = cdexpr ->
	  <:expr< Rtcg.float_bin $e1$ $f$ $e2$ >>
      ] 

    | "*" LEFTA
      [ e1 = cdexpr;
        f = [ "*"    -> <:expr< Rtcg.IOpMUL >> 
            | "/"    -> <:expr< Rtcg.IOpDIV >> 
	    | "land" -> <:expr< Rtcg.IOpAND >>
	    | "lor"  -> <:expr< Rtcg.IOpOR >>
	    | "lxor" -> <:expr< Rtcg.IOpXOR >>   
	    | "mod"  -> <:expr< Rtcg.IOpMOD >> 
            ];
	e2 = cdexpr ->
	  <:expr< Rtcg.int_bin $e1$ $f$ $e2$ >>
      | e1 = cdexpr;
	f = [ "*." -> <:expr< Rtcg.FLOP_MUL >>
            | "/." -> <:expr< Rtcg.FLOP_DIV >> 
	    | "%." -> <:expr< Rtcg.FLOP_MOD >> ];
	e2 = cdexpr ->
	  <:expr< Rtcg.float_bin $e1$ $f$ $e2$ >>
      ] 

    | "**" LEFTA
      [ e1 = cdexpr;
        f = [ "lsl" -> <:expr< Rtcg.IOpLSL >> 
            | "lsr" -> <:expr< Rtcg.IOpLSR >> 
	    | "asr" -> <:expr< Rtcg.IOpASR >>
            ];
	e2 = cdexpr ->
	  <:expr< Rtcg.int_bin $e1$ $f$ $e2$ >>

      | e1 = cdexpr; "**"; e2 = cdexpr ->
	  <:expr< Rtcg.float_bin $e1$ Rtcg.FLOP_POWER $e2$ >>
      ] 

    | "unary minus" NONA
      [ "-"; e = cdexpr ->
	<:expr< Rtcg.int_un Rtcg.IUnNEG $e$ >>
      | "lnot"; e = cdexpr ->
	<:expr< Rtcg.int_un Rtcg.IUnNOT $e$ >>  
      |	"-."; e = cdexpr ->
	<:expr< Rtcg.float_neg $e$ >>
      ]

    | "apply" LEFTA 
      [ f = SELF; a = SELF ->
	  <:expr< Rtcg.apply $f$ $a$ >>
      | "ref"; e = cdexpr -> 
	  <:expr< Rtcg.ref_create $e$ >> 

      |	u = UIDENT; ca = cdexpr_constr ->
	  fun_of_constr_arg (CAConstr(u,ca,loc))

      |	u = UIDENT -> <:expr< Rtcg.quote $uid:u$ >>
      ]
    | "." LEFTA
      [ e1 = cdexpr; "."; "("; e2 = cdexpr; ")" ->
	  <:expr< Rtcg.array_get $e1$ $e2$ >>      
      | e1 = cdexpr; "."; "("; e2 = cdexpr; ")"; "<-"; v = cdexpr ->
	  <:expr< Rtcg.array_set $e1$ $e2$ $v$ >>
      |	e1 = cdexpr; "."; "["; e2 = cdexpr; "]" ->
	  <:expr< Rtcg.string_get $e1$ $e2$ >>      
      | e1 = cdexpr; "."; "["; e2 = cdexpr; "]"; "<-"; v = cdexpr ->
	  <:expr< Rtcg.string_set $e1$ $e2$ $v$ >>
      |	e1 = SELF; "."; l = LIST1 value_path SEP "." ->
	  let body =  
	    List.fold_left 
	      (fun expr qid -> 
		 let eqid = expr_of_value_path loc qid in
		 <:expr< $expr$ . $eqid$ >> )
	      <:expr< r >> l in
	  <:expr< Rtcg.apply (Rtcg.quote (fun r -> $body$)) $e1$ >>
        
      ] 
    | "~-" NONA
      [
        "!"; e = cdexpr -> <:expr< Rtcg.ref_value $e$ >>
      | "`"; e = expr LEVEL "simple"-> <:expr< Rtcg.quote $e$ >> 
      | "@"; e = expr LEVEL "simple" -> e
      ] 
    | "simple" 
      [
      	"true" -> <:expr< Rtcg.quote True >>
      |	"false" -> <:expr< Rtcg.quote False >>
      |	"^"; e = cdexpr -> <:expr< Rtcg.lose $e$ >>   
      | n = INT -> <:expr< Rtcg.quote $int:n$ >> 
      | n = FLOAT -> <:expr< Rtcg.quote $flo:n$ >> 
      |	s = STRING -> <:expr< Rtcg.quote $str:s$ >>
      | qid = value_path -> expr_of_value_path loc qid
      | "("; e = cdexpr; ")" -> e 
      | "["; "]" -> <:expr< Rtcg.quote [] >>
      | "["; el = cdexpr1_semi_list; "]" -> 
	  gen_list_creation loc el
      | "{"; test_label_eq; lbl = cd_lbl_list; "}" ->
	  gen_record_creation loc None lbl
      | "{"; e = cdexpr LEVEL "simple"; "with"; lbl = cd_lbl_list; "}" ->
	  gen_record_creation loc (Some e) lbl
      ]
  ];

cd_lbl_list:
  [
    [ le = cd_lbl; ";"; lel = cd_lbl_list -> le::lel
    | le = cd_lbl; ";" -> [le]
    | le = cd_lbl -> [le]
    ]
  ];

cd_lbl:
  [ [ qid = value_path; "="; e = cdexpr LEVEL "expr1" -> (qid, e)
    ]
  ];

cdexpr_constr:
  [  
     "tuple"
     [
       "("; l = LIST0 SELF SEP ","; ")" -> 
	 CATuple l
     ]  
  |  "constr"
     [
       u = UIDENT; c = cdexpr_constr LEVEL "tuple" ->
	 CAConstr(u,c,loc)
     | u = UIDENT ->
	 CAConstrNull(u,loc)
     ] 
  |  
     [ e = cdexpr LEVEL ":=" -> CAExpr e ] 
  ];

cd_match_case:
  [
    [ 
      x1 = patt; w = OPT [ "when"; e = expr -> e ]; 
         "->"; x2 = cdexpr -> (x1, w, x2)
    ]
  ];
    

value_path:
  [[
    lid = LIDENT -> ([], lid)
  | uid = UIDENT; "."; (modules,lid) = value_path -> (uid::modules, lid)
  ]];

cdexpr1_semi_list:
  [
    [ e = cdexpr LEVEL "expr1"; ";"; el = SELF -> e :: el
    | e = cdexpr LEVEL "expr1"; ";" -> [e]
    | e = cdexpr LEVEL "expr1" -> [e]
    ]
  ];

END;; 


(*
 * $Log: dml.ml4,v $
 * Revision 1.8  2002/04/29 11:00:16  mitya
 * Added record creation
 *
 * Revision 1.7  2002/03/15 17:14:39  mitya
 * *** empty log message ***
 *
 * Revision 1.6  2002/03/15 10:14:58  mitya
 * List creation (as in [x;y;z]) added.
 *
 * Revision 1.5  2002/03/15 09:00:37  mitya
 * Pattern-matching generation added
 *
 * Revision 1.4  2002/03/14 19:25:05  mitya
 * Bug fixed in definition of "::"
 *
 * Revision 1.3  2002/03/12 12:50:30  mitya
 * Record field accesses added.
 *
 * Revision 1.2  2002/03/01 11:25:36  mitya
 * Constructors added to syntax
 *
 *)
